package DBIx::Class::Storage::DBI::ODBC::DB2_400_SQL;

use strict;
use warnings;

use base qw/
    DBIx::Class::Storage::DBI::ODBC
    DBIx::Class::Storage::DBI::DB2
/;
use mro 'c3';

1;

=head1 NAME

DBIx::Class::Storage::DBI::ODBC::DB2_400_SQL - Support specific to DB2/400
over ODBC

=head1 DESCRIPTION

This is an empty subclass of L<DBIx::Class::Storage::DBI::DB2>.

=head1 AUTHOR

See L<DBIx::Class/AUTHOR> and L<DBIx::Class/CONTRIBUTORS>.

=head1 LICENSE

You may distribute this code under the same terms as Perl itself.

=cut
# vim:sts=2 sw=2:
